
#include "dht11.h"
#include "jackalope.h"

#if DHT11_ENABLE == 1

#define DHT11_IO    PD7
#define DHT11_PORT  PORTD
#define DHT11_PIN   PIND
#define DHT11_DDR   DDRD
#define DHT11_DEBUG 0
// Sensor Type DHT-11 ord DHT-22 dont use DHT11 its totally unprecise
#define DHT_SENSOR  22

void dht11_read(int16_t *temp, int16_t *hum, uint8_t *err)
{
    *err = 1;
    uint16_t i;
    uint8_t old_val, new_val;
    uint64_t result  = 0;

    uint8_t h_high;
    uint8_t h_low;
    uint8_t t_high;
    uint8_t t_low;
    uint8_t parity;
    uint8_t bitcount = 0;

    //int16_t act_temp;
    //int16_t act_hum;

#if DHT11_DEBUG == 1
    uint8_t bitcount = 0;
    uint16_t bits[43] = {0};
#endif

    // 1. Ausgang auf LOW ziehen und warten
    DHT11_DDR |= _BV(DHT11_IO);
    DHT11_PORT &= ~_BV(DHT11_IO);

    // _delay_loop_2 = 4 CPU Cycles / count = 5 count = 1us @ 20MHz
    // 20ms warten
    for(i=0; i<20; i++)
    {
        _delay_loop_2(5000); // 1ms
    }

    // Ausgang wieder als eingang schalten mit pullup an
    DHT11_DDR &= ~_BV(DHT11_IO); // Eingang
    DHT11_PORT |= _BV(DHT11_IO); // pullup an

    // Eingang pollen
    old_val = 0;
    i = 0;
    while (i < 1000)
    {
        i++;
        new_val = DHT11_PIN & _BV(DHT11_IO);
        if (old_val != new_val)
        {
            if (new_val > 0)
            {
                i=0;
            }
            else
            {
                if (i > 20)
                { result |= (uint64_t)0x01;}
                else
                { }
                result = (result << 1);

                //DEBUG
#if DHT11_DEBUG == 1
                bits[bitcount]=i;
                i=0;
                if (bitcount == 42) { i= 10000; }
#endif
                bitcount++;
            }
            old_val = new_val;
        }
        _delay_loop_2(5);
    }


#if DHT11_DEBUG == 1
    for(i=0;i<43;i++)
    {
        debug_printf("BIT: %u %d  ",i,bits[i]);
    }
#endif

    result = (result >> 1);

    parity = (uint8_t)result;
    result = (result >> 8);
    t_low = (uint8_t)result;
    result = (result >> 8);
    t_high = (uint8_t)result;
    result = (result >> 8);
    h_low = (uint8_t)result;
    result = (result >> 8);
    h_high = (uint8_t)result;

    uint8_t p;
    p =  t_low+t_high+h_low+h_high;
    if (p != parity)
    {
        *err = 7;
    }

    if (bitcount < 40)
    {
        *err = 8;
    }

#if DHT_SENSOR == 11
    *temp = (int16_t)t;
    *hum = (int16_t)h;
#endif
#if DHT_SENSOR == 22
    uint16_t t;
    //debug_printf("HI: %u\r\n",t_high);
    //debug_printf("LO: %u\r\n",t_low);
    //debug_printf("TEMP: %u\r\n",t);
    t = t_high*256+t_low;
    if (t > 32768)
    {
        *temp = (int16_t)t - 32768;
        *temp = *temp * -1;
    }
    else
    {
        *temp = t;
    }
    *hum = (int16_t)h_high*256+h_low;
    //debug_printf("TEMP: %d C HUM: %d PARITY: 0x%02x",act_temp,act_hum,parity);
#endif
}

#endif
