#include "eeprom.h"

static void eeprom_write(uint16_t, uint8_t );
static uint8_t eeprom_read(uint16_t);
static uint16_t crc_checksum(uint8_t *,uint8_t);

uint16_t eeprom_write_bytearray(uint8_t *bytes,uint16_t len, uint16_t offset)
{
    /* crc berechnen */
    uint16_t crc = crc_checksum(bytes,len);
    uint16_t i;

    len += offset;
    for (i=offset; i<len; i++)
    {
        eeprom_write(i,bytes[(i-offset)]);
    }
    eeprom_write(i++,LO8(crc));
    eeprom_write(i,HI8(crc));

    /* Lesen und vergleichen */
    for (i=offset; i<len; i++)
    {
        if (bytes[(i-offset)] != eeprom_read(i))
        {
            return i;
        }
    }

    if (eeprom_read(i++) != LO8(crc))
    {
        return i;
    }

    if (eeprom_read(i) != HI8(crc))
    {
        return i;
    }

    return 0;
}

uint16_t eeprom_read_bytearray(uint8_t *bytes, uint16_t len, uint16_t offset)
{
    uint16_t i;
    for (i=offset; i<(len+offset); i++)
    {
        bytes[(i-offset)] = eeprom_read(i);
    }
    // crc berechnen
    uint16_t crc = crc_checksum(bytes,len);

    if (eeprom_read(i++) != LO8(crc))
    {
        return 0;
    }

    if (eeprom_read(i) != HI8(crc))
    {
        return 0;
    }

    return i-(offset+1);
}

void eeprom_read_raw_bytearray(uint8_t *bytes, uint16_t len, uint16_t offset)
{
    uint16_t i;
    for (i=offset; i<(len+offset); i++)
    {
        bytes[(i-offset)] = eeprom_read(i);
    }
}

void eeprom_write(uint16_t address, uint8_t data)
{
    /* Warten bis das vorhergende write abgeschlossen ist */
    while( ( EECR & _BV(EEPE) ) > 0)
    {
        asm("nop");
    }

    /* Adresse und Daten setzen */
    EEAR = address;
    EEDR = data;

    /* schreiben starten */
    EECR |= _BV(EEMPE);
    EECR |= _BV(EEPE);
}

uint8_t eeprom_read(uint16_t address)
{
    /* Warten bis das vorhergende write abgeschlossen ist */
    while( ( EECR & _BV(EEPE) ) > 0)
    {
        asm("nop");
    }

    /* Adresse setzen */
    EEAR = address;

    /* lesen starten */
    EECR |= _BV(EERE);

    return EEDR;
}


uint16_t crc_checksum(uint8_t *data, uint8_t length)
{
    uint16_t crc = 0;
    uint8_t *p = (uint8_t *)data;

    for (uint8_t i = 0; i < length; i++) {
        crc = _crc16_update(crc, *p);
        p++;
    }
    return crc;
}
