#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <util/delay.h>
#include <util/crc16.h>

#define HI8(x)  ((uint8_t)((x) >> 8))
#define LO8(x)  ((uint8_t)(x))

#define FNORD_RX_BUFLEN = 20;

// Definiert in jackalope.h
//#define FNORD_NET_DEBUG 1
//#define FNORD_NET_DEBUG_DATA 0

/* fnord net initialisation */
extern void fnord_init(uint8_t);

/* Fnordframe prcessing for RX and TX needs to be called often (every 100ms for example)*/
extern void fnord_process(void);

/* check if a setup frame has been successfully transmitted to the destination 1=yes 0=no */
/* uint8_t fnord_check_setup(uint8_t destination); */
extern uint8_t fnord_check_setup(uint8_t);

/* before you can send data to a destination, you need to initialize the destination */
/* this function will just compse a frame transmit is done by fnord_process() */
/* void fnord_transmit_setup(uint8_t send_to) */
extern void fnord_transmit_setup(uint8_t);

/* transmit a fnord frame, this function will just compse a frame transmit is done by fnord_process() */
/* void fnord_transmit_frame(uint8_t send_to, uint8_t *payload, uint8_t payloadlength) */
extern void fnord_transmit_frame(uint8_t,uint8_t *,uint8_t);

/* silence counter management */
extern void fnord_countup_silence_count(void);
extern void fnord_clear_silence_count(void);

/* Transmission in progress flag management */
extern void set_fnord_tx_in_progress(uint8_t);

/* set local address */
void fnord_set_local_address(uint8_t);

/* get status for last transmitted frame 0=undef, 1=TX OK, 10=TX fail */
uint8_t fnord_get_tx_status(void);

// see fnord_check_frame for details
uint8_t fnord_get_rx_status(void);

/* enduser receive for fnordframe data */
uint8_t fnord_get_rx_datalen(void);

 // clears rx_datalen returns a pointer to rx_data array
uint8_t *fnord_get_rx_data(void);

// when fnord_net wants to send data it calls this function
// this function has to be implemented by the main programm
// which includes fnord-net.h
// parameters: pointer to uint8_t data array , datalength in bytes
void fnord_tx_request(uint8_t *,uint8_t);

// before fnordnet transmits data it can check for on air collissions
// listening somehow if somebody ele transmits something and waiting till
// this transmission is over, this has to be implemented by the RF module driver
// retuns 0 for transmission ok, 1 for medium busy
uint8_t fnord_check_collision(void);

// This function is to be called when a frame was received from
// a datasource and needs to bes passed to fnord-net protocol
void fnord_received_frame(uint8_t *,uint8_t);

uint32_t crc32_checksum(void *,uint8_t);
