#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>

#include "rfm69.h"
#include "fnord-net.h"
#include "jackalope.h"

#if RFM69_ENABLE == 1

/* Gobale Variablen RFM69 */
static uint8_t rfm69_rx_data[RFM69_RX_BUFLEN];

static uint8_t rfm69_tx_data[RFM69_TX_BUFLEN];
static uint8_t rfm69_tx_len;

static uint32_t volatile rfm69_irqcnt;
static uint8_t volatile rfm69_gotirq;

/* RFM69 Parameters */
// Channels with 75 KHz Bandwidth
//TODO richtige werte eintragen
//const uint16_t PROGMEM rfm69_channels[23] = {1238,1268,1298,1328,1358,1388,1418,1448,1478,1508,1538,1568,1598,1628,1658,1688,1718,1748,1778,1808,1838,1868,1898};
//const uint32_t PROGMEM rfm69_channels[23] = {7095828,7097057,7098286,7099514,7100743,7101972,7103201,7104430,7105658,7106887,7108116,7109345,7110574,7111802,7113031,7114260,7115489,7116718,7117946,7119175,7120404,7121633,7122862};

const uint32_t PROGMEM rfm69_channels[23] = {433095000,433170000,433245000,433320000,433395000,433470000,433545000,433620000,433695000,433770000,433845000,433920000,433995000,434070000,434145000,434220000,434295000,434370000,434445000,434520000,434595000,434670000,434745000};

void rfm69_init(void)
{
    // Deselect RFM69 chip
    DDRB |= _BV(PB1);
    PORTB |= _BV(PB1);

    rfm69_rc_calibration(); // For listen mode not currently used
    rfm69_setmodulation(0); // 0=FSK 1=OOK
    rfm69_setmode(0); // Packet mode
    rfm69_setdeviation(30); // 30 KHZ FSK frequency deviation
    //rfm69_setdeviation(15); // 15 KHZ FSK frequency deviation

    rfm69_set_rxtx_mode(0); // Enable Receive Mode and set DIO0 assignment
    rfm69_setreg(0x11,0x80); // Power Amp + Output Power select

    //TODO
    rfm69_setrxbw(0); // derzeit fest 60 kHz

    // addr, mask, data
    rfm69_setregmask(0x18,0x80,0x00); // LNA Zin = 50 Ohm Bit 7 = 0
    //rfm69_setregmask(0x18,0x80,0x80); // LNA Zin = 200 Ohm Bit 7 = 1

    rfm69_setreg(0x2c,0x00); // Preamble (0xAA) count MSB = 0
    rfm69_setreg(0x2d,0x03); // Preamble (0xAA) count LSB = 3

    rfm69_setreg(0x2e,0x88); // UseSyncWord=ON, SyncWordLen=2Byte, FillFifo when SYNC Match
    rfm69_setreg(0x2f,0x2d); // Sync Word MSB
    rfm69_setreg(0x30,0xd4); // Sync Word LSB
    //rfm69_setreg(0x2f,0xc3); // Sync Word MSB
    //rfm69_setreg(0x30,0x83); // Sync Word LSB

    rfm69_setreg(0x37,0x80); //VarLenPacket DontCalcCRC // Seite 54
    //rfm69_setreg(0x37,0x00); //FixedLenPacket DontCalcCRC
    //rfm69_setreg(0x38,0x20); // Payloadlength MAX 32 Byte Seite 52 factory default 64 Byte

    rfm69_setreg(0x3c,0x04); // Start TX by FIFO Threshold 4 BYTE -> FifoLevelInterrupt Seite 73 / 46

    rfm69_setreg(0x3d,0x00); // RX Auto Restart OFF Seite 73

    // Auto mode for Transmit
    // Seite 42 Auto mode Register 0x3b 0x01
    //rfm69_setreg(0x3b,0x5b); // Enter FIFO Level, Exit Packet Send
    rfm69_setreg(0x3b,0x00); // Auto Modes off

    // 433 MHz Test
    //rfm69_setmode(3); // Continous without sync
    //rfm69_setmodulation(1); // 0=FSK 1=OOK
    //DIO0 RSSI Indicator
    //rfm69_setreg(0x25,0x80);

    //rfm69_setreg(0x25,0xC0); // RSSI in packet mode

    // Inter Packet RX Delay ??
    // RegPacketConfig2 0x3d ??

    // Enable Pinchange Interrupt
    DDRC &= ~_BV(PC5); // PC5 is input
    PORTC &= ~_BV(PC5); // Pullup Resistor OFF
    PCICR |= _BV(PCIE1); // Enable Pin change Interrupt 1
    PCMSK1 |= _BV(PCINT13); // On Pin PC5 = PCINT 13
}

// According to page 40 of datasheet RC Oscillator needs to be calibrated at power up
void rfm69_rc_calibration(void)
{
    uint16_t i = 0;
    rfm69_setreg(0x0A,0x80);
    while ( (rfm69_getreg(0x0A) & 0x40) == 0 && i < 10000)
    {
        asm("nop");
        i++;
    }
    if (i > 9999)
    {
        debug_printf("RFM69 seems to be dead\r\n");
    }
    else
    {
        debug_printf("RFM69 communication OK\r\n");
    }
}

void rfm69_interrupt_routine(void)
{
    //433MHz OOK Tx/Rx ongoing nothing todo for FnordNet
    //debug_printf("RFM69 Irq %d\r\n",rfm_mode);
    uint8_t pinlevel;
    pinlevel = ((PINC & 0b00100000) >> 5);

    //debug_printf("IRQ %d %d\r\n",rfm_mode,pinlevel);

    //uint8_t status = rfm69_getreg(0x28);
    //debug_printf("Reg: 0x%02x",status);

    // Rising edge only
    if ( pinlevel == 0)
    { return; }

    if (rfm_mode > 9)
    {
        if (rfm_mode == 20)
        {
            rfm_mode = 19;
        }

        // OOK RX Fifo ready
        if (rfm_mode == 18)
        {
            rfm_mode = 17;
        }
        return;
    }

    // TX Mode
    if (rfm_mode == 1)
    {
        // PacketSend  ??
        if ( ( rfm69_getreg(0x28) & 0x08 ) > 0 )
        {
            rfm69_set_rxtx_mode(0);
#if FNORD_NET_ENABLE == 1
            set_fnord_tx_in_progress(0);
            fnord_clear_silence_count();
#endif
            return;
        }
    }

    // RX Mode
    if (rfm_mode == 0)
    {
        // Standby modus damit das RSSI Flag in Reg 0x27 geloescht wird
        rfm69_set_rxtx_mode(2);

        //rfm69_read_fifo(rfm69_rx_data,20); // Debug received data

        // Frame(FIFO) nach rfm69_rx_data kopieren
        rfm69_read_fifo(rfm69_rx_data,RFM69_RX_BUFLEN);
        rfm69_received_frame(rfm69_rx_data,rfm69_rx_data[0]);

        // setze nach standby wieder in den receive modus
        rfm69_set_rxtx_mode(0);

        // 433MHz OOK test
        //DIO0 ist RSSI Indicator DIO2 ist data
        //rfm69_setreg(0x25,0x80);

        // RSSI in Packet mode
        //rfm69_setreg(0x25,0xC0);
    }
    rfm69_irqcnt++;
}



void rfm69_setbaudrate(uint16_t baud)
{
    float divisor_f = (float) 32000000UL / baud;
    uint16_t divisor = (uint16_t) (divisor_f + 0.5);
    uint8_t lowbyte = (uint8_t) divisor & 0xff;
    uint8_t highbyte = (uint8_t) (divisor >> 8);
    // LFLAGS += -Wl,-u,vfprintf -lprintf_flt -lm //in makefile aktivieren damit float ausgegeben wird
    //debug_printf("RFM69 SETBAUD LOW:0x%02x HIGH: 0x%02x FLOAT: %-.2f\r\n",lowbyte,highbyte,divisor_f);
    rfm69_setreg(0x03,highbyte); // Baudrate highbyte
    rfm69_setreg(0x04,lowbyte); // Baudrate lowbyte
}

// Output Power in 32 Steps (0-31) -18 dbm to +13 dBm in 1 db steps
int16_t rfm69_settxpower(uint8_t power)
{
        int16_t retval = -18;
        if (power > 31 ) { power = 31;}
        uint8_t val = rfm69_getreg(0x11);
        val &= 0x80;
        val += power;
        rfm69_setreg(0x11,val);
        return retval+power;
}

// Frequenzhub für FSK in kHz //
void rfm69_setdeviation(uint8_t fdev)
{
    if (fdev > 125) {fdev = 125;}
    uint32_t value = fdev * 1000;
    uint16_t reg = value / 61;
    uint8_t reg_lo = (uint8_t) reg & 0xff;
    uint8_t reg_hi = (uint8_t) (reg >> 8);
    rfm69_setreg(0x05,reg_hi);
    rfm69_setreg(0x06,reg_lo);
}

/* Traeger Frequenz einstellen
 *
 * 433,05 MHz - 434,79 MHz
 */
void rfm69_setfrequency(uint32_t freq)
{
    float reg_val_f;
    uint32_t reg_val;
    uint8_t reg_lo;
    uint8_t reg_mid;
    uint8_t reg_hi;

    //float step = 61,03515625f ; // frequenzstufe  = 32000000f / 524288f

    reg_val_f = (float)freq / 61.03515625f;
    reg_val_f += 0.5;
    reg_val = (uint32_t)reg_val_f;

    reg_lo = (uint8_t)(reg_val & 0xff);
    reg_mid = (uint8_t)(reg_val >> 8);
    reg_hi = (uint8_t)(reg_val >> 16);

    //debug_printf("RFM69 SETFREQ LOW:0x%02x MED: 0x%02x HIGH: 0x%02x\r\n",reg_lo,reg_mid,reg_hi);
    //debug_printf("RFM69 SETFREQ LOW:%d MED:%d HIGH:%d F: %ld\r\n",reg_lo,reg_mid,reg_hi,freq);
    rfm69_setreg(0x07,reg_hi);
    rfm69_setreg(0x08,reg_mid);
    rfm69_setreg(0x09,reg_lo);
}

void rfm69_setchannel(uint8_t newchannel)
{
    if (newchannel > 0 && newchannel < 24)
    {
        uint8_t channel = --newchannel;
        uint32_t f0 = pgm_read_dword(&rfm69_channels[channel]);
        rfm69_setfrequency(f0);
    }
}

// TODO Seite 26
void rfm69_setrxbw(uint16_t bw)
{
    //uint8_t exp = 2;
    //uint8_t mant = 0; // wert 0 = 16(125 KHz), wert 8 = 20(100 KHz), wert 16 = 24(83 KHz)

    // Formel: RxBw = 32000000 / ( (mant*4 + 16) * 2 ^ (exp+2) )

    uint8_t mant = 0; // mantisse bit 3 und 4 (8 und 16) werte 0,1,2   3=reserviert
    uint8_t exp = 4; // Exponent = untere 3 bit werte 0-7

    // Wertetabelle (Auszug)
/*
     Mant Exp BW (MHz)
        2   6   5.2
        1   6   6.3
        0   6   7.8
        2   5   10.4
        1   5   12.5
        0   5   15.6
        2   4   20.8
        1   4   25.0
        0   4   31.3
        2   3   41.7
        1   3   50.0
        0   3   62.5
        2   2   83.3
*/
    mant = (mant << 3);

    uint8_t val = rfm69_getreg(0x19);
    val = val & 0xe0;
    val += (exp+mant);
    rfm69_setreg(0x19,val);

}

uint32_t rfm69_get_irqcount(void)
{
    return rfm69_irqcnt;
}

uint8_t rfm69_get_chiprevision(void)
{
    return (rfm69_getreg(0x10) >> 4);
}

/* Not used
uint8_t rfm69_rssi_scan(uint16_t anz)
{
    uint8_t rssi;
    uint8_t rssi_threshold;
    rssi_threshold = rfm69_getreg(0x29);
    rfm69_setreg(0x29,240); // min rssi = -120dBm
    rssi =  rfm69_min_rssi(anz);
    rfm69_setreg(0x29,rssi_threshold);
    return rssi;
}
*/




uint8_t rfm69_getrssi(void)
{
    uint8_t val = 0;
    // Start RSSI Scan //
    rfm69_setreg(0x23,0x01);
    // Warten bis RSSI SCAN fertig ist
    while ( (rfm69_getreg(0x23) & 0x02) == 0 )
    {
    }
    val = rfm69_getreg(0x24);
    //debug_printf("RSSI Result: %u\r\n",val);
    val = val / 2;

    return val;
}


/* this function scans for the noise level */
uint8_t rfm69_min_rssi(int16_t testcycles, int16_t *min_val, int16_t *max_val)
{
    uint16_t i = 0;
    uint8_t val = 0;
    uint32_t result = 0;
    uint8_t min = 0;
    uint8_t max = 255;
    for (i = 0 ; i < (uint16_t)testcycles ; i++)
    {
        val = rfm69_getrssi();
        if (val > min) { min = val; }
        if (val < max) { max = val; }
        result = result + val;
        //debug_printf("Result: %d dB\r\n",val);
    }
    *min_val = min;
    *max_val = max;
    //debug_printf("Min: -%u dBm Max: -%u dBm\r\n",min,max);
    val = result / testcycles ;
    return val;
}

void rfm69_setrssi(uint8_t rssi)
{
    //if ( rssi >= 0 && rssi < 125) // rssi >= 0 ist ummer true für unsigned int
    if ( rssi < 125)
    {
        rssi = rssi * 2;
        rfm69_setreg(0x29,rssi);
    }
}

void rfm69_tx_bytestream(uint8_t *bytestream,uint8_t bytecount)
{
    // Set Standyby mode
    rfm69_set_rxtx_mode(2);

    // Fill FIFO Sync Byte = 2D D4 set in rfm69_init()
    rfm69_tx_data[0] = bytecount + 2; // incl. null byte das als letztes uebertragen wird RFM69 Seite 52

    uint8_t i;
    for (i = 0 ; i< bytecount; i++)
    {
        rfm69_tx_data[i+1] = bytestream[i];
    }

    rfm69_tx_data[bytecount + 2] = 0x55;
    rfm69_tx_len = bytecount + 2;

    //debug_printf("TX: LEN: %d\r\n",rfm69_tx_len);

    // Write data to fifo
    rfm69_write_fifo(rfm69_tx_data, rfm69_tx_len);

    // Set TX Mode
    rfm69_set_rxtx_mode(1);

#if FNORD_NET_ENABLE == 1
            // last clear in rfm69 interrupt routine
            fnord_clear_silence_count();
#endif

    // Interrupt behandelt wenn der Versand fertig ist

    // Warten bis der Versand fertig ist
    /*
    i=0;
    while ( i < 10)
    {
        rfm69_getreg(0x28);
        asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop");
        i++;
    }
    */
}

/* Modulation einstellen 0=FSK 1=OOK */
void rfm69_setmodulation(uint8_t modtype)
{
    uint8_t val = rfm69_getreg(0x02);
    modtype &= 0x01;
    modtype = (modtype << 3);
    val &= 0xF7;
    val |= modtype;
    rfm69_setreg(0x02,val);

    // Gaussian filter BT 1 = gerinsgter effekt 0.3 = stärkster effekt
    if (modtype == 0)
    {
        // gaussian filter BT = 1.0
        val = rfm69_getreg(0x02);
        val &= 0xFC;
        val |= 0x01;
        rfm69_setreg(0x02,val);
    }
    else if (modtype == 1)
    {
        // Kein filter fuer OOK
        val = rfm69_getreg(0x02);
        val &= 0xFC;
        //val |= 0x00;
        rfm69_setreg(0x02,val);
    }
}

/* 0 = packet mode 1=reserved 2=continious mode with bit sync 3= cont mode without bit sync */
void rfm69_setmode(uint8_t mode)
{
    uint8_t val = rfm69_getreg(0x02);
    mode &= 0x03;
    mode = (mode << 5);
    val &= 0x9F;
    val |= mode;
    rfm69_setreg(0x02,val);
}

void rfm69_set_rxtx_mode(uint8_t rxtx)
{
    cli(); // no interrupts while switching modes
    rfm_mode = rxtx;
    // RX Mode
    uint8_t val;
    if (rxtx == 0)
    {
        rfm69_setreg(0x25,0x40); // DIO0 Mapping RX=PayloadReady  Seite 48
        val = rfm69_getreg(0x01);
        val &= 0xE3;
        val |= 0x10;
        rfm69_setreg(0x01,val);
        //while ((rfm69_getreg(0x27) & 0x80) > 0) {}
    }

    //TX Mode
    if (rxtx == 1 || rxtx == 20)
    {
        rfm69_setreg(0x25,0x00); // DIO0 Mapping TX=PacketSend  Seite 48
        val = rfm69_getreg(0x01);
        val &= 0xE3;
        val |= 0x0C;
        rfm69_setreg(0x01,val);
        //while ((rfm69_getreg(0x27) & 0x80) > 0) {}
    }

    // Standby
    if (rxtx == 2)
    {
        val = rfm69_getreg(0x01);
        val &= 0xE3;
        val |= 0x04;
        rfm69_setreg(0x01,val);
        //while ((rfm69_getreg(0x27) & 0x80) > 0) {}
    }
    sei();
}


/*
uint8_t rfm69_getstatus()
{
    * RegIrqFlags1 (0x27)
}
*/

uint8_t rfm69_getreg(uint8_t addr)
{
    return rfm69_sendcmd(addr,0,0);
}

uint8_t rfm69_setreg(uint8_t addr, uint8_t data)
{
    return rfm69_sendcmd(addr,data,1);
}

uint8_t rfm69_setregmask(uint8_t addr,uint8_t mask,  uint8_t data)
{
    uint8_t val = rfm69_getreg(addr);
    mask = ~mask;
    val = val & mask;
    val = val + data;
    return rfm69_sendcmd(addr,val,1);
}

uint8_t rfm69_sendcmd(uint8_t addr, uint8_t data, uint8_t readwrite)
{
    uint8_t tmp = 0;
    uint8_t val = 0;

    addr |= (readwrite << 7);

    /* Chipselect auf low */
    DDRB |= _BV(PB1);
    PORTB &= ~_BV(PB1);

#if SPI_CMD_DEBUG == 1
    debug_printf("RFM69 SPI CMD: REG: 0x%02x SEND DATA: 0x%02x %d\r\n",addr,data,data);
#endif
    tmp = spi_transfer_byte(addr);
    val = spi_transfer_byte(data);

    PORTB |= _BV(PB1);

#if SPI_CMD_DEBUG == 1
    debug_printf("RFM69 SPI CMD: REG: 0x%02x  GOT DATA: 0x%02x %d\r\n",addr,val,val);
#endif
    return val;
}

void rfm69_read_fifo(uint8_t *data, uint8_t len)
{
    uint8_t tmp = 0;
    uint8_t i;

    /* Chipselect auf low */
    DDRB |= _BV(PB1);
    PORTB &= ~_BV(PB1);

    // Read fifo Command
    tmp = spi_transfer_byte(0x00);

    // Read fifo Command
    for (i=0; i<len ; i++)
    {
        data[i] = spi_transfer_byte(0);
        //debug_printf("RFM69 FIFO %d val: %d\r\n",i,data[i]);
    }

    // Durch das setzen von RX nach Standby in rfm69_interrupt_routine
    // und danach wieder in den RX modus, wird der FIFO komplett geloscht

    /* Chipselect auf high */
    PORTB |= _BV(PB1);
}

void rfm69_write_fifo(uint8_t *data, uint8_t len)
{
    uint8_t tmp = 0;
    uint8_t i;

    /* Chipselect auf low */
    DDRB |= _BV(PB1);
    PORTB &= ~_BV(PB1);

    // Write fifo Command
    tmp = spi_transfer_byte(0x80);

    for (i=0; i<len ; i++)
    {
        tmp = spi_transfer_byte(data[i]);
        //debug_printf("RFM69 WRITE FIFO %d,0x%02x\r\n",i,data[i]);
    }

    /* Chipselect auf high */
    PORTB |= _BV(PB1);
}
#endif
